////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//// eslint
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// globals from common.js - override .eslintrc.json defaults
/* globals unifiedCreateObjectURL:true unifiedRevokeObjectURL:true logs_collections:true */

// globals from hooks
/* globals unified_URL */

// globals from hooks
/* globals unified_URL */

// misc js libs
/* global CryptoJS psl pHash OcrFactory $ */

// browser apis
/* global browser:true */

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//// polyfills
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*
	Object.assign() polyfill taken from MDN 
	https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/assign#Polyfill
*/
if (typeof Object.assign != 'function') {
	// Must be writable: true, enumerable: false, configurable: true
	Object.defineProperty(Object, "assign", {
		value: function assign(target, varArgs) { // .length of function is 2
			'use strict';
			if (target == null) { // TypeError if undefined or null
				throw new TypeError('Cannot convert undefined or null to object');
			}

			var to = Object(target);

			for (var index = 1; index < arguments.length; index++) {
				var nextSource = arguments[index];

				if (nextSource != null) { // Skip over if undefined or null
					for (var nextKey in nextSource) {
						// Avoid bugs when hasOwnProperty is shadowed
						if (Object.prototype.hasOwnProperty.call(nextSource, nextKey)) {
							to[nextKey] = nextSource[nextKey];
						}
					}
				}
			}
			return to;
		},
		writable: true,
		configurable: true
	});
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//// alter build-in prototypes
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// custom format function
Date.prototype._SBA4B_format = function (format) {
	var o = {
		"M+": this.getMonth() + 1, //month
		"d+": this.getDate(),    //day
		"h+": this.getHours(),   //hour
		"m+": this.getMinutes(), //minute
		"s+": this.getSeconds(), //second
		"q+": Math.floor((this.getMonth() + 3) / 3),  //quarter
		"S": this.getMilliseconds() //millisecond
	};

	if (/(y+)/.test(format)) format = format.replace(RegExp.$1,
		(this.getFullYear() + "").substr(4 - RegExp.$1.length));
	for (var k in o) if (new RegExp("(" + k + ")").test(format))
		format = format.replace(RegExp.$1,
			RegExp.$1.length == 1 ? o[k] :
				("00" + o[k]).substr(("" + o[k]).length));
	return format;
};

// String.prototype.endsWith polyfill for IE - taken from MDN
// extends the String prototype with a private property. that is to prevent collision
// with alternative "endsWith" implementations provided by pages we inject into
String.prototype._SBA4B_endsWith = function(searchStr, Position) {
	// This works much better than >= because
	// it compensates for NaN:
	if (!(Position < this.length))
		Position = this.length;
	else
		Position |= 0; // round position
	return this.substr(Position - searchStr.length, searchStr.length) === searchStr;
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//// utility function
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

let Utils = {
	assert: function(expression, message) {
		if (!expression) throw new Error(message);
	},

	getTwoWayMap: function(arr) {
		let twoWayMap = new Map();
		arr.forEach(function(e) { 
			twoWayMap.set(e[0],e[1]); 
			twoWayMap.set(e[1],e[0]); 
		});
		return twoWayMap;
	},

	lazy: function(func) {
		let retVal;
		return (function() {
			if (!retVal) {
				retVal = func.apply(this, arguments);
			}
			return retVal;
		});
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//// abstract definitions - should have browser specific implementations in hooks js files
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

let load_hooks_impl = function(){};

let funcs = {};
funcs.add_on_message_listener		= function(func){ Logger.debug(Logger.modules.MAIN, mainLogSessionId, "add_on_message_listener - empty implementation"); };
funcs.set_badge_text				= function(text){ Logger.debug(Logger.modules.MAIN, mainLogSessionId, "set_badge_text - empty implementation"); };
funcs.set_badge_icon 				= function(icon){ Logger.debug(Logger.modules.MAIN, mainLogSessionId, "set_badge_icon - empty implementation"); };
funcs.get_app_version				= function(full){ Logger.debug(Logger.modules.MAIN, mainLogSessionId, "get_app_version - empty implementation"); return "N/A Version"; };
funcs.get_build_number				= function(){ Logger.debug(Logger.modules.MAIN, mainLogSessionId, "get_build_number - empty implementation"); return "N/A Version"; };
funcs.get_distribution_id			= function(){ Logger.debug(Logger.modules.MAIN, mainLogSessionId, "get_distribution_id - empty implementation"); return "N/A Version"; };
funcs.get_installation_url			= function(callback){Logger.debug(Logger.modules.MAIN, mainLogSessionId, "get_installation_url - empty implementation"); return undefined; };
funcs.is_tab_open					= function(url, activate, callback){Logger.debug(Logger.modules.MAIN, mainLogSessionId, "is_tab_open - empty implementation"); return undefined; };
funcs.notifications_create			= function(id,type,progress,icon,title,message,buttons,clearBefore,durationSeconds){ Logger.debug(Logger.modules.MAIN, mainLogSessionId, "notifications_create - empty implementation"); };
funcs.notifications_clear			= function(id){ Logger.debug(Logger.modules.MAIN, mainLogSessionId, "notifications_clear - empty implementation"); };
funcs.load_history 					= function() { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "load_history - empty implementation"); };  // required to fill browserHistory map. 
funcs.history_deleteUrl				= function(url) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "history_deleteUrl - empty implementation"); };
funcs.history_onVisited_addListener = function(callback) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "history_onVisited_addListener - empty implementation"); };
funcs.get_resource_url				= function(resource_name)  { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "get_resource_url - empty implementation"); };
funcs.storage_remove_key			= function(key) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "storage_remove_key - empty implementation"); };
funcs.storage_store					= function(key,value,func)  { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "storage_store - empty implementation"); };
funcs.storage_store_all				= function(elem,func) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "storage_store_all - empty implementation"); };
funcs.storage_managed_load			= function(policy) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "storage_managed_load - empty implementation"); };
funcs.storage_load					= function(key,func)  { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "storage_load - empty implementation"); };
funcs.cancelDownload 				= function(download_id) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "cancelDownload - empty implementation"); };
funcs.capture_screen				= function(entry,func) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "capture_screen - empty implementation");  func(entry); };
funcs.send_message					= function(message) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "send_message - empty implementation"); };
funcs.usercheck_popup				= function(page, entry){ Logger.debug(Logger.modules.MAIN, mainLogSessionId, "usercheck_popup - empty implementation"); };
funcs.usercheck_password_reuse_popup= function(url, tab_id, index){ Logger.debug(Logger.modules.MAIN, mainLogSessionId, "usercheck_password_reuse_popup - empty implementation"); };
funcs.register_error_events 		= function() { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "register_error_events - empty implementation"); };
funcs.register_events				= function() { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "register_events - empty implementation"); };
funcs.get_current_url 				= function(cb) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "get_current_url - empty implementation"); };
funcs.download_file 				= function(donwloadDone_cb,params) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "download_file - empty implementation"); };
funcs.suggestFileName				= function(id, params, fileInfo) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "suggestFileName - empty implementation"); };
funcs.handle_open_file				= function(downloadId) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "handle_open_file - empty implementation"); };
funcs.handle_show_folder			= function(downloadId) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "handle_show_folder - empty implementation"); };
funcs.handle_open_downloads			= function() { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "handle_open_downloads - empty implementation"); };
funcs.handle_open_options  			= function() { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "handle_open_options - empty implementation"); };
funcs.get_favicon					= function(url, callback) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "get_favicon - empty implementation"); };
funcs.get_extension_name			= function() { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "get_extension_name - empty implementation"); };
funcs.create_tab					= function(tab, func) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "create_tab - empty implementation"); };
funcs.update_tab					= function(tab_id, url) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "update_tab - empty implementation"); };
funcs.update_iframe					= function(tab_id, origin_url, block_page) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "update_iframe - empty implementation"); };
funcs.close_tab						= function(tab, func) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "close_tab - empty implementation"); };
funcs.close_current_tab				= function(func) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "close_current_tab - empty implementation"); };
funcs.tab_send_message 				= function(id, message, options, func) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "tab_send_message - empty implementation"); };
funcs.handle_get_orig				= function(name, url, quarantine) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "handle_get_orig - empty implementation"); };
funcs.create_context_menu			= function() { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "create_context_menu - empty implementation"); };
funcs.set_log_level					= function(level) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "set_log_level - empty implementation"); };
funcs.get_locale_message			= function(message) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "get_locale_message - empty implementation"); return message; };
funcs.get_sandblast_file			= function(path) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "get_sandblast_file - empty implementation"); return ""; };
funcs.get_collected_log				= function(path) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "get_collected_log - empty implementation"); return ""; };
funcs.detect_language				= function(tabId, cb) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "detect_language - empty implementation"); cb("en"); };
funcs.get_cookie					= function(url, xhr) { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "get_cookie - empty implementation"); };
funcs.trust_server_cert				= function() { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "trust_server_cert - empty implementation"); };
funcs.all_tabs_update_config		= function() { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "all_tabs_update_config - empty implementation"); };
funcs.browser_name					= function() { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "browser - empty implementation"); return ""; };
funcs.try_get_user_id				= function() { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "try_get_user_id - empty implementation"); };
funcs.get_manifest					= function() { Logger.debug(Logger.modules.MAIN, mainLogSessionId, "get_manifest - empty implementation"); return {}; };

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//// URLF constants
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// possible "connection check" modes
const CHECK_CONN_MODE_GATEWAY = "gateway";
const CHECK_CONN_MODE_TE_CLOUD = "cloud";
const CHECK_CONN_MODE_LOG_SERVER = "log_server";

// possible threat prevention protocols
const TP_PROTOCOL_CLOUD = "cloud";
const TP_PROTOCOL_GAIA = "gaia";

// possible TE server configurations
const TE_SERVER_MODE_GATEWAY = "gateway";
const TE_SERVER_MODE_CLOUD = "cloud";
const TE_SERVER_MODE_FALLBACK_CLOUD = "fallback_to_cloud";

// constant related to handling of "Office Editing for Docs, Sheets & Slides" extension
const officeEditingDistributionIds = {
	'gbkeegbaiigmenfmjfclcdgdpimamgkj': 1,
	'bpmcpldpdmajfigpchkicefoigmkfalc': 1
};

const officeMime2FileExtension = {
	"application/msword": "doc",
	"application/vnd.ms-word": "doc",
	"application/vnd.msword": "doc",
	"application/vnd.openxmlformats-officedocument.wordprocessingml.document": "docx",
	"application/vnd.wordprocessing-openxml": "docx",
	"application/vnd.ms-word.document.macroenabled.12": "docm",
	"application/vnd.ms-word.document.12": "docm",
	"application/vnd.openxmlformats": "docx",
	"application/msword-template": "dot",
	"application/vnd.openxmlformats-officedocument.wordprocessingml.template": "dotx",
	"application/mspowerpoint": "ppt",
	"application/vnd.ms-powerpoint": "ppt",
	"application/vnd.openxmlformats-officedocument.presentationml.presentation": "pptx",
	"application/vnd.presentation-openxml": "pptx",
	"application/vnd.presentation-openxmlm": "pptx",
	"application/vnd.ms-powerpoint.presentation.macroenabled.12": "pptm",
	"application/vnd.openxmlformats-officedocument.presentationml.template": "potx",
	"application/msexcel": "xls",
	"application/vnd.ms-excel": "xls",
	"application/vnd.openxmlformats-officedocument.spreadsheetml.sheet": "xlsx",
	"application/vnd.spreadsheet-openxml": "xlsx",
	"application/vnd.ms-excel.sheet.macroenabled.12": "xlsm",
	"application/vnd.openxmlformats-officedocument.spreadsheetml.template": "xltx",
};

const officeFileExtension2Mime = {
	"doc": "application/msword",
	"docx": "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
	"docm": "application/vnd.ms-word.document.macroenabled.12",
	"dot": "application/msword-template",
	"dotx": "application/vnd.openxmlformats-officedocument.wordprocessingml.template",
	"xls": "application/vnd.ms-excel",
	"xlsx": "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
	"xlsm": "application/vnd.ms-excel.sheet.macroenabled.12",
	"xlt": "application/vnd.ms-excel",
	"xltx": "application/vnd.openxmlformats-officedocument.spreadsheetml.template",
	"csv": "text/csv",	// csv is not supported by SBA4B, but can be still opened by locally by the "Office Editing" extension
	"ppt": "application/vnd.ms-powerpoint",
	"pptx": "application/vnd.openxmlformats-officedocument.presentationml.presentation",
	"pptm": "application/vnd.ms-powerpoint.presentation.macroenabled.12",
	"pot": "application/vnd.ms-powerpoint",
	"potx": "application/vnd.openxmlformats-officedocument.presentationml.template"
};

// all dictionary of store ids used to deploy K12. used to determine if URLF UI should be enabled by default given no GPO was deployed
const K12_distributionIds = {
	"hjggfiblbamimgahoclhpmkfkndgcmgo" : 1,
	"jjhgpiidbafdbdfdbllombimlkoloccl" : 2,
	"bpajiabahopiegjodmjlbejeiedehigp" : 3
};

// URLF categories, used by the UI
var categories = {
	45: ["Alcohol & Tobacco", "This category includes URLs that are related to the sale, promotion, or consumption of alcoholic or tobacco (tobacco related) products. This also includes sites that contain cocktail recipes and home brewing instructions. Examples: http://www.wine.com, http://www.smokers.com, http://www.finckcigarcompany.com"],
	52000038: ["Anonymizer", "An intermediary which prevents Web sites from seeing a user's Internet Protocol (IP) address. Anonymizers are often used to circumvent company network acceptable use policies to potentially hide non-appropriate behavior, however in the home setting they may be used to provide additional layers of anonymity and protect privacy. Examples: http://www.anonymizer.com, http://www.megaproxy.com, http://www.hidemyass.com"],
	47: ["Art / Culture", "This category includes URLs that contain virtual art galleries, artist sites (includes sculpture and photograph), and other cultural topics like museums, ethnic customs, and country/regional customs. URLs that contain potentially offensive material will also be categorized with the appropriate category (e.g. Nudity). Examples: http://www.polkmuseumofart.org, http://www.matsuoka-museum.jp"],
	65: ["Botnets", "Sites that use bots (zombies) including command-and-control sites."],
	51: ["Business / Economy", "This category includes URLs that provide business-related information. This includes corporate overviews, business planning, and strategies as well as information, services, or products that help other businesses plan, manage, or market their enterprises. Examples: http://www.cat.com, www.cint.com, www.brandsoftheworld.com"],
	52: ["Chat / Instant Messaging", "Ability to make a connection to other clients and send Instant Messages through a web or client side interface. Examples: http://www.icq.com, http://www.tawk.to"],
	1: ["Computers / Internet", "This category is intended to cover websites related to computing software and hardware, as well as Internet and technology-related companies. This includes, but is not limited to vendors, product reviews, and deployment and maintenance of software and hardware. This also includes addons such as scripts, plugins, drivers, peripherals, and other equipment used in conjunction with computers and networks. Examples: http://www.archive.org, http://www.verisign.com, http://www.limewire.com, http://www.w3schools.com"],
	3: ["Education", "This category includes home pages of colleges, universities, and K-12 schools. Examples: http://www.wikipedia.org, http://www.berkeley.edu, http://www.biology-online.org"],
	52000130: ["Email", "Webmail is e-mail that is accessed via a web browser. Examples: http://mail.live.com, http://mail.yahoo.com, http://www.gmail.com"],
	5: ["Entertainment", "This category includes URLs that provide information about the entertainment industry. This includes, but is not limited to, cinema, theater, music, television, infotainment, entertainment industry gossip, sites about actors and musicians (including fan clubs), and horoscopes. Examples: http://www.hollywood.com, http://disney.go.com, http://www.cinema-source.com/cinemasource/media.htm"],
	68: ["Fashion", "Sites concerning fashion. Includes product reviews, comparisons, and general consumer information."],
	52000136: ["File Storage and Sharing", "This category includes file sharing applications and sites, for public or private sharing of data or space in a network. The data or space can be used, opened, changed, or deleted according to access privilege."],
	7: ["Financial Services", "This category includes URLs that provide financial information or access to online financial accounts. This includes, but is not limited to personal or business related banking, money management, tax consulting, mutual funds, credit cards, credit unions, insurance, or other financial publications and services. Examples: http://www.marketwatch.com, http://www.paypal.com, http://www.bankofamerica.com/index.cfm, http://www.etrade.com"],
	9: ["Gambling", "This category includes URLs that allows users to place wagers or bets online or provide gambling-related software. This includes, but is not limited to online casinos, betting pools, sports betting, lotteries, and sites that are relevant to such activities. Examples: http://www.silverdollar.com, http://www.desertdollar.eu, http://www.tnlottery.com"],
	52000058: ["Games", "A game is an activity that provides entertainment or amusement. A game can be played solo or with a large group of people. Games can be played with or without structured rules depending on the game. Games generally challenge the mental and physical prowess of the person who is gaming. Examples: http://videogames.yahoo.com, http://www.megagames.com"],
	11: ["Government / Military", "This category includes URLs with content maintained by governmental organizations (e.g. government branches or agencies, police departments, fire departments, military branches or armed services) and supranational organizations such as the United Nations or the European Community. Sites maintained by non-governmental entities that provide information about government, military branches, or armed services are also included. Examples: http://www.ca.gov, http://www.army.mil, http://www.gc.ca0"],
	12: ["Greeting Cards", "This category includes URLs that allow people to send and receive digital greeting cards via the Internet. These sites are used in the same way that people use conventional greeting cards (e.g. birthdays or holidays). They present a risk as sensitive or proprietary information may be leaked, or messages could contain objectionable information, malware, or harmful file attachments. Examples: http://www.americangreetings.com"],
	13: ["Hacking", "This category includes URLs that provide how-to information or otherwise enable fraud, crime, or computer-oriented malicious activity. This includes, but is not limited to sites that contain root kits, scripts or other software that can be used to compromise other computers, or otherwise illegally manipulate electronic devices."],
	14: ["Hate / Racism", "This category includes URLs that provide information that encourage oppression of a specific group of people. This includes, but is not limited to promoting (either explicitly or implicitly) an agenda against groups based on race, religion, nationality, gender, age, disability, or sexual orientation. This category also includes political parties that have a hate-based agenda. Examples: http://www.kkkk.net"],
	15: ["Health", "This category includes URLs that cover health-related information and health care services. This includes, but is not limited to physical or mental wellbeing, diet, nutrition, fitness, parenting, health insurance, hospitals, clinics, and Physicians. Examples: http://www.kidshealth.org, http://www.dietitian.com, http://www.webmd.com"],
	17: ["Illegal / Questionable", "This category includes URLs that either directly distribute or provide information on obtaining illegally distributed software, music, or film. This also includes, but is not limited to, illegal license key generators, software cracks, and serial numbers. Examples: http://www.cheathouse.com"],
	18: ["Illegal Drugs", "This category includes URLs that provide information on the purchase, manufacture, and use of illegal or recreational drugs. This includes, but is not limited to, information related to drug paraphernalia (displaying, selling, or detailing the use of), tips on legal highs (e.g. using non drugs or misusing legal drugs), abuse of other legal substances and any Marijuana related sites. Example: http://www.hightimes.com"],
	67: ["Inactive Sites", "Sites that are inactive, typically reserved for later use. They most often do not contain their own content, may simply say 'under construction', 'purchase this domain', or display advertisements."],
	50000086: ["Instant Chat", "Ability to make a connection to other clients and send Instant Messages through a web or client side interface. Examples: http://www.icq.com, http://www.tawk.to"],
	52000047: ["Instant Messaging", "Also called Chat Programs. These applications are installed on a PC by a user, and allow immediate chats between other users of the program. Some of these programs also allow audio or video conferencing as well as the ability to execute file transfers."],
	21: ["Job Search / Careers", "This category includes URLs related to finding a job or changing careers. This can include, but is not limited to resume writing, interviewing techniques, corporate web pages listing job openings, classified job advertising, and large job databases. Examples: http://www.careerone.com.au, http://www.careerbuilder.com"],
	52000051: ["Media Sharing", "Applications used to combine audio and visual media for the purpose of communicating. Multimedia applications are largely used in entertainment and advertising. Examples: http://www.flickr.com, http://www.webshots.com, http://images.search.yahoo.com"],
	23: ["Media Streams", "This category includes URLs that provide streaming audio or video, e.g. Internet radio, Internet TV stations, or Podcasting. This category is intended to cover audio and/or video that use significant amounts of bandwidth. Audio or Video streams with potentially offensive material will also be categorized with the relevant category (e.g. Intimate Apparel/Swimsuit, Nudity, Tasteless). Examples: http://www.napster.com, http://www.youtube.com, http://www.mp3.com"],
	24: ["News / Media", "This category includes URLs that provide online news media such as International or regional news broadcasting and publication. Examples: http://www.cnn.com, http://www.nytimes.com, http://www.washingtonpost.com"],
	25: ["Newsgroups / Forums", "This category includes URLs that provides access to Usenet news or other forums for online exchange of ideas. The category is meant to cover the risks associated with content that changes frequently according to user activity as it may range from harmless to offensive. Furthermore, these sites are often not highly trafficked, making it difficult to categorize and provide appropriate coverage for each page. Examples: http://forums.offtopic.com, http://forums.clubsi.com, http://groups.google.com"],
	69: ["Non-profits & NGOs", "Sites devoted to clubs, communities, unions, and non-profit organizations. Many of these groups exist for educational or charitable purposes. Examples: http://www.advocateweb.org"],
	26: ["Nudity", "This category includes URLs that include non-pornographic images of the bare human body like those that may be found in classic sculpture and paintings or medical images. Examples: http://www.imageofmuse.com, http://www.met-art.com"],
	52000046: ["P2P File Sharing", "Short for Peer-to-Peer. A peer-to-peer (or P2P) computer network is a network that relies primarily on the computing power and bandwidth of the participants in the network rather than concentrating it in a relatively low number of servers. P2P networks are typically used for connecting nodes via largely ad hoc connections. Such networks are useful for many purposes but have different impact on end users and the Enterprise. File sharing, containing audio, video, data or anything in digital format is very common as well as realtime data, such as telephony traffic, is also passed using P2P technology. Examples: http://www.imesh.com"],
	30: ["Personals / Dating", "This category includes URLs that provide networking for the purpose of online dating, match-making, escort services, or introductions to potential spouses. These sites allow for the exchange of personal information and facilitate real-time communication. These sites may allow contact with undesirable parties and member profiles may contain objectionable information. Note that some social networking sites also include dating functions, but it is not their primary focus. Those sites will not be included in the Personals/Dating category. Examples: http://www.match.com, http://www.dating.com, http://www.loveme.com"],
	31: ["Phishing", "This category includes URLs that typically arrive in emails and are established only to steal information from users. These sites falsely represent themselves as the website of a legitimate company in order to obtain user account information that can be used for fraudulent and/or illegal purposes."],
	32: ["Political / Activist Groups", "This category includes URLs covering political parties, individuals in political life, and opinion on various topics. It also includes sites that offer legal information and advice. Examples: http://www.texasgop.org/site/PageServer, http://www.antifaschistische-nachrichten.de"],
	33: ["Pornography", "This category includes URLs that contain materials that are intended to be sexually arousing or erotic. This includes fetish pages, animation, cartoons, stories, and child pornography.Examples: xvideos.com, sex.com."],
	34: ["Real Estate", "This category includes URLs that provide information and services on commercial or residential real estate. This includes, but is not limited to sales or rental of living or retail space, guides for apartments, housing, and property, as well as appraisal and brokerage information. Examples: http://www.greatlandrealty.com, http://www.coldwellbanker.com"],
	35: ["Recreation", "This category includes URLs related to recreational activities and hobbies. This includes, but is not limited to recreational facilities such as swimming pools, zoos, fairs, amusement parks, recreation guides, hiking, fishing, bird watching, and stamp collecting. Examples: http://adoption.about.com, http://www.sandiegozoo.org, http://www.vix.com"],
	37: ["Religion", "This category includes URLs that pertain to religious topics and human spirituality. This includes, but is not limited to mainstream and fringe religions and groups, and spiritual conjecture and philosophy. Examples: http://www.isgh.org, http://www.scientology.org"],
	39: ["Restaurants / Dining / Food", "This category includes URLs that provide information about restaurants, bars, catering, delivery and takeout. This includes, but is not limited to information about restaurants, online ordering of food, restaurant guides, and reviews. Examples: http://www.chevys.com , http://www.starbucks.com"],
	52000132: ["Search Engines / Portals", "A search engine is a commonly used web application designed to let the user search for content based on the terms used. Examples: http://www.yahoo.com, http://www.google.com, http://search.yahoo.com"],
	70: ["Sex", "This category includes URLs that contain sex related content such as pornography or nudity. Examples: http://www.hustler.com, http://www.eroticwriting.co.uk"],
	41: ["Sex Education", "This category includes URLs that provide descriptions or depictions of sexual acts, but is primarily intended for educational purposes and not intended to be arousing or erotic. Examples: http://www.sexualhealth.com"],
	42: ["Shopping", "This category includes URLs that specifically sell products or services or advertise the means by which products or services can be obtained. Many online shipping sites pose a risk to users by offering access to items that would normally be filtered under other categories such as Pornography, Weapons, or Violence. Examples: http://www.ebay.com, http://amazon.com"],
	52000069: ["Social Networking", "Social networking is a method to link communities of people together in some fashion, usually by software. The software, though varies from network to network usually gives users the ability to share and upload photos, videos and bookmarks, blog, message friends, link with or make new friends. Examples: http://www.linkedin.com, http://www.facebook.com, http://www.myspace.com"],
	53: ["Software Downloads", "This category includes URLs that provide legal downloadable software for personal computers. Not all software available on the Internet is appropriate for use inside a secured network. The category is meant to cover the risks associated with downloading and installing unapproved software inside the secure network. Examples: http://www.top-of-software.de, http://www.download.cnet.com"],
	66: ["Spam", "Sites that have been promoted through spam techniques."],
	54: ["Sports", "This category includes URLs related to sports, whether professional or recreational. This includes sporting news, events, and information such as playing tips, strategies, game scores, or player trades. Examples: http://www.trails.com, http://www.nfl.com"],
	55: ["Spyware / Malicious Sites", "This category includes URLs that deploy code specifically designed to hijack you computer activity or change the settings on your computer to undesirable or less secure settings. This can include, but is not limited to self-installing applications (a.k.a. Drive-by executable file downloads), Trojan horses, and viruses that exploit security vulnerabilities in browsers or other applications."],
	56: ["Tasteless", "This category includes URLs that contain content that can be considered tasteless, gross, or gruesome. This includes, but is not limited to tasteless humor, bodily excretory functions, graphic photos that contain wounds or blood, extreme body modification (cutting, branding, genital piercing), or graphic depictions of inhumane animal treatment."],
	58: ["Translation", "This category includes URLs that allow users to enter a block of text or other URLs to translate the content from one language to another. Also included are language identifiers. These sites may present a risk that confidential information may be leaked or exposed to offensive words. Examples: http://translate.google.com/translate, http://babelfish.yahoo.com"],
	59: ["Travel", "This category includes URLs designed to promote travel for either business or personal reasons. This includes, but is not limited to information about hotels, resorts, airlines, ground transportation, car rentals, travel agencies, and general travel information. This also includes sites that allow purchase of tickets or secure accommodations. Travel sites streams with potentially offensive material will also be categorized with the relevant category (e.g. Intimate Apparel/Swimsuit, Nudity). Examples: http://www.expedia.com, http://www.flysaa.com, http://www.polhotels.com"],
	0: ["Blacklisted", "Blacklisted"],
	60: ["Vehicles", "This category includes URLs for manufacturers and dealerships of transportation vehicles such as cars, vans, trucks, SUVs, motorcycles and scooters. This category also includes sites that provide marketing, reviews, comparisons, pricing, and general information about motor vehicles. Examples: http://www.chryslerllc.com, http://steeringnews.com/ , http://www.downtownhd.com"],
	61: ["Violence", "This category includes URLs that contain real or lifelike images or text that portray, describe, or advocate physical assaults against humans, animals, or institutions (for example: depictions of war, suicide, mutilation, dismemberment). Examples: http://www.2violent.com"],
	62: ["Weapons", "This category includes URLs that provide information about weapons either for military or personal use. This can include, but is not limited to purchasing, making, modifying, or using weapons and includes both lethal weapons like guns, knives, or swords, and non-lethal weapons like paintball guns. Ammunition, explosives, and weapon accessories are also included in this category. Examples: http://www.gunaccessories.com, http://www.zmweapons.com, http://www.galleryofguns.com"],
	50000032: ["Web Advertisements", "Displays third-party advertising in pop-ups, pop-unders, interstitials, sliders, daughter consoles and any other ?windows? that the software opens on the users' desktops which are not part of internet content that users directly choose to visit or view. Examples: http://www.exoclick.com , http://www.adclicks-agent.de"],
	51000004: ["High Risk", "Sites used to host malware content (e.g. spyware, phishing and malicious sites) or site that potentially host such malicious content. Sites that are identified as definite malicious will be re-categorized to one of Malicious category accordingly. Sites that are identified as definite clear sites will be re-categorized to one of the non-malicious category accordingly."],
	49: ["Blogs / Personal Pages", "This category includes URLs where users post or edit content quickly and spontaneously. This includes, but is not limited to blogs and wikis. The category is meant to cover the risks associated with content that changes frequently according to user activity as it may range from harmless to offensive."],
	73: ["Child Abuse", "Sites that portray or discuss children in sexual or other abusive acts."],
	51000005: ["Critical Risk", "Applications and Websites that may bypass security or hide identity."],
	71: ["General", "Sites that do not clearly fall into other categories, for example, blank web pages."],
	75: ["Lifestyle", "This category includes URLs that provide information about lifestyle. This includes, but is not limited to hobbies, leisure activities, community involvement and other subjects that comprise the way of life."],
	76: ["Lingerie and Swimsuit / Suggestive", "Sites that offer images of models in suggestive costume, with semi nudity permitted. Includes sites offering lingerie or swimwear. Examples: http://www.microbikiniswimsuits.net, http://www.aquawearswim.com"],
	51000002: ["Low Risk", "Applications and Websites that are potentially non business related yet low risk."],
	78: ["Marijuana", "This category includes URLs that contain information, discussions, or sale of Marijuana and associated products or services, including legalizing marijuana and/or using marijuana for medicinal purposes. See sk112186 for more information. Example: http://purethc.com"],
	51000003: ["Medium Risk", "Applications and Websites that may be misused and cause data leak / malware infection."],
	72: ["Nature / Conservation", "Sites related to green energy, ecology, nature and environment."],
	77: ["Suspicious Content", "Sites used to host malware content (e.g. spyware, phishing and malicious sites) or site that potentially host such malicious content. Sites that are identified as definite malicious will be re-categorized to one of Malicious category accordingly. Sites that are identified as definite clear sites will be re-categorized to one of the non-malicious category accordingly."],
	51000001: ["Very Low Risk", "Applications and Websites that are usually business related with no or very low risk."]
};

// default categories enforcement
const block_lvl_high = [45, 65, 9, 13, 14, 17, 18, 26, 31, 33, 70, 66, 55, 61, 62, 37, 51000003, 51000004, 51000005, 52000038];
const block_lvl_medium = [45, 65, 9, 13, 14, 17, 18, 26, 31, 33, 70, 66, 55, 61, 62, 51000004, 51000005];
const block_lvl_low = [9, 13, 14, 17, 18, 26, 31, 33, 70, 66, 55, 61, 51000005];
const default_bullying_enforced = [ 52000069, 52000047, 50000086, 52 ];

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//// globals & constants
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

const ExtensionDistEnum = Object.freeze({
	"SBA": 1,
	"ZONE_ALARM": 2,
	"K12": 3,
	"ORANGE_X": 4
});


const reputationService = "https://rep.checkpoint.com/Phishing";

// node
const node_domain = "127.0.0.1";
const node_web_server_is_https = true;
const node_web_server_port = 8443;
const node_server = (node_web_server_is_https ? "https" : "http") + "://" + node_domain + ":" + node_web_server_port; // + "/phishing"; 

const file_protection_mode_flags = {
	"default": 0,
	"allow" : 1,
	"block" : 2,
	"extract" : 4,
	"emulate" : 8,
	"both" : 12
};

const threat_emulation_flags = {
	"default": 0,
	"background" : 1,
	"wait" : 2
};

const threat_extraction_flags = {
	"profile" : 0,
	"clean" : 1,
	"pdf" : 2
};

const threat_extraction_flags_reversed = getReverseMap(threat_extraction_flags);

const use_te_cloud_flags = {
	"false" : 0,
	"true" : 1,
	"fallback_to_cloud" : 2
};

const threat_extraction_parts_codes = [
	"sensitive_hyperlinks",
	"macros_and_code",
	"embedded_objects",
	"linked_objects",
	"database_queries",
	"pdf_launch_app",
	"pdf_play_audio",
	"pdf_play_video",
	"pdf_open_url",
	"pdf_execute_code",
	"pdf_submit_form",
	"pdf_go_to_remote",
	"custom_properties",
	"fast_save_data",
	"statistic_properties",
	"summary_properties", 
	"three_d_artwork_annotations",
	"embedded_images"
];

const keys_description = {
	"sensitive_hyperlinks" 	:	{"subject" : "Extracted Elements", "gw_value" : 1034, "default": true, "desc" : "Sensitive Hyperlinks", "full_desc"  : "Links to network/local file paths"},
	"macros_and_code"	 	:	{"subject" : "Extracted Elements", "gw_value" : 1026, "default": true, "desc" : "Macros and Code", "full_desc"  : "Microsoft Office macros and PDF JavaScript code"},
	"embedded_objects"	 	:	{"subject" : "Extracted Elements", "gw_value" : 1019, "default": true, "desc" : "Embedded Objects", "full_desc"  : "Files and objects embedded in documents"},
	"linked_objects"	 	:	{"subject" : "Extracted Elements", "gw_value" : 1025, "default": true, "desc" : "Linked Objects", "full_desc"  : "Linked objects"},
	"database_queries"	 	:	{"subject" : "Extracted Elements", "gw_value" : 1018, "default": true, "desc" : "Database Queries", "full_desc"  : "Queries to remote databases"},
	"pdf_launch_app"	 	:	{"subject" : "Extracted Elements", "gw_value" : 1139, "default": true, "desc" : "PDF Launch Actions", "full_desc"  : "Launch external applications"},
	"pdf_play_audio"	 	:	{"subject" : "Extracted Elements", "gw_value" : 1142, "default": true, "desc" : "PDF Sound Actions", "full_desc"  : "Play sound objects"},
	"pdf_play_video"	 	:	{"subject" : "Extracted Elements", "gw_value" : 1143, "default": true, "desc" : "PDF Movie Actions", "full_desc"  : "Play movie files"},
	"pdf_open_url"		 	:	{"subject" : "Extracted Elements", "gw_value" : 1141, "default": true, "desc" : "PDF URI Actions", "full_desc"  : "Open Uniform Resource Identifier (URI) resources"},
	"pdf_execute_code"	 	:	{"subject" : "Extracted Elements", "gw_value" : 1150, "default": true, "desc" : "PDF JavaScript Actions", "full_desc"  : "Execute JavaScript code"},
	"pdf_submit_form"	 	:	{"subject" : "Extracted Elements", "gw_value" : 1151, "default": true, "desc" : "PDF Submit Form Actions", "full_desc"  : "Submit data to remote locations"},
	"pdf_go_to_remote"	 	:	{"subject" : "Extracted Elements", "gw_value" : 1137, "default": true, "desc" : "PDF GoToR Actions", "full_desc"  : "Open other PDF files"},
	"custom_properties"	 	:	{"subject" : "Extracted Elements", "gw_value" : 1017, "default": false, "desc" : "Custom Properties", "full_desc"  : "Custom document properties"},
	"fast_save_data"	 	:	{"subject" : "Extracted Elements", "gw_value" : 1021, "default": true, "desc" : "Fast Save Data", "full_desc"  : "Stored data for fast document saving"},
	"statistic_properties"	:	{"subject" : "Extracted Elements", "gw_value" : 1036, "default": false, "desc" : "Statistic Properties", "full_desc"  : "Statistic document properties"},
	"summary_properties"	:	{"subject" : "Extracted Elements", "gw_value" : 1037, "default": false, "desc" : "Summary Properties", "full_desc"  : "Summary document properties"},
	"three_d_artwork_annotations":	{"subject" : "Extracted Elements", "gw_value" : 1178, "default": false, "desc" : "3D Artwork Annotations", "full_desc"  : "PDF 3D Artwork Annotations"},
	"embedded_images"		:	{"subject" : "Extracted Elements", "gw_value" : 500,  "default": false, "desc" : "Embedded Images", "full_desc"  : "Cleans images embedded in documents"}
};

const browser_name = {
	"ie" : 0,
	"chrome" : 1,
	"firefox_legacy" : 2,
	"firefox_quantum" : 3,
	"edge" : 4,
	"node" : 5
};

const default_flags_for_docs = calculateFileProtectionFlags(file_protection_mode_flags['both'], threat_emulation_flags['wait'], threat_extraction_flags['clean']);
const default_flags_for_other_supported = calculateFileProtectionFlags(file_protection_mode_flags['emulate'], threat_emulation_flags['wait'], threat_extraction_flags['profile']);
const default_flags_for_unsupported = calculateFileProtectionFlags(file_protection_mode_flags['allow'], threat_emulation_flags['wait'], threat_extraction_flags['profile']);

const default_flags_for_docs_K12 = calculateFileProtectionFlags(file_protection_mode_flags['extract'], threat_emulation_flags['wait'], threat_extraction_flags['clean']);
const default_flags_for_unsupported_K12 = calculateFileProtectionFlags(file_protection_mode_flags['allow'], threat_emulation_flags['wait'], threat_extraction_flags['profile']);
const default_flags_for_docs_orangex = calculateFileProtectionFlags(file_protection_mode_flags['extract'], threat_emulation_flags['wait'], threat_extraction_flags['clean']);
const default_flags_for_unsupported_orangex = calculateFileProtectionFlags(file_protection_mode_flags['allow'], threat_emulation_flags['wait'], threat_extraction_flags['profile']);


// log levels
const LOG_LVL_AUTOMATION = 5;
const LOG_LVL_DEBUG = 4;
const LOG_LVL_INFO = 3;
const LOG_LVL_WARN = 2;
const LOG_LVL_ERROR = 1;

var mainLogSessionId = getRandomToken(4); //for logs

const collect_logs_enabled = true;

let debug, info, warn, error;	// will hold logger function references
let logs_collections = [];
var AutomationLogger = { // for selenium automation
	"collect_logs": function() {
		funcs.send_message({"command": "auto_collect_new_logs"});
	},
	"new_logs": "",
	"new_logs_flag": 0,
	"new_logs_index": 0,
	"send_logs": function(tabID) {
		AutomationLogger["new_logs"] = get_collected_log().slice(AutomationLogger["new_logs_index"], -1);
		AutomationLogger["new_logs_index"] += AutomationLogger["new_logs"].length;
		funcs.tab_send_message(tabID, {"command": "auto_send_new_logs", "logs": AutomationLogger["new_logs"]});
	}
};

// define universal object URL handling functions based on browser. workaround the issue of Webkit based browsers providing different API for URL
let unifiedCreateObjectURL = function() {};
let unifiedRevokeObjectURL = function() {};

(function defineUnifiedObjectURLHandlers() {
	if ((typeof URL !== 'undefined') && URL.createObjectURL) {
		unifiedCreateObjectURL = URL.createObjectURL;
		unifiedRevokeObjectURL = URL.revokeObjectURL;
	}
	else if ((typeof window !== 'undefined') && window.URL && window.URL.createObjectURL) {
		unifiedCreateObjectURL = window.URL.createObjectURL;
		unifiedRevokeObjectURL = window.URL.revokeObjectURL;	
	}
	else if ((typeof window !== 'undefined') && window.webkitURL) {
		unifiedCreateObjectURL = window.webkitURL.createObjectURL;
		unifiedRevokeObjectURL = window.webkitURL.revokeObjectURL;
	}
})();

// define a global 'browser' object depending on browser, but don't overwrite if it's already defined (firefox quantum)
if (typeof window !== "undefined" && typeof browser === "undefined") {
	var browser;	// global scope!
	browser = window.msBrowser || window.browser || window.chrome;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//// function definitions
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// random token related functions
function getRandomToken(bytes) { // E.g. 16 * 8 = 128 bits token
    var randomPool = new Uint8Array(bytes);
    getRandomValues(randomPool);
    var hex = '';
    for (var i = 0; i < randomPool.length; ++i) {
		var x = randomPool[i].toString(16);
        hex += (x.length == 1) ? ("0" + x) : x;
    }
    return hex;
}

function isRandomToken(str, bytes) {
	if (!str || (str.length != (bytes * 2)))
		return false;
	return /^[0-9A-Fa-f]+$/g.test(str);
}

// logger object
let Logger = {
	levels: {
		ERROR: 1,
		WARN: 2,
		INFO: 3,
		DEBUG: 4,
		AUTOMATION: 5,
		TRACE: 6
	},
	levelToText: {
		1: "ERROR",
		2: "WARN",
		3: "INFO",
		4: "DEBUG",
		5: "AUTOMATION",
		6: "TRACE"
	},
	modules: {
		DL_PROTECTION: "FILE_DOWNLOAD",
		ZERO_PHISHING: "ZERO_PHISHING",
		ANTIBULLYING: "ANTI_BULLYING",
		PASS_REUSE: "PASSWORD_REUSE",
		URLF: "URLF",
		MAIN: "MAIN",
		UI_NOTIF: "UI_NOTIF",
		GOOGLE_ANALITICS: "ANALITICS"
	},
	
	maxLogLevel: 5,	// default max log level

	/**
	 * set maximal log level
	 * @param {number} logLevel
	 */
	setMaxLogLevel: function(logLevel) {
		Utils.assert(Logger.levelToText[logLevel], 'invalid log level');
		Logger.maxLogLevel = logLevel;
	},

	/**
	 * trim message string to length
	 * @param {string} msg message string
	 * @returns trimmed message string
	 */
	trimMsg: function(msg) {
		const maxMsgLength = 200;
		if (msg.length > maxMsgLength) {
			msg = msg.substr(0,maxMsgLength) + '...';
		}
		return msg;
	},

	/**
	 * escape message for logger. if empty string, return placeholder
	 * @param {string} msg message string
	 * @param {boolean} shouldTrim if true, will trim the message string to length
	 * @returns escaped message string
	 */
	escapeMsg: function (msg, shouldTrim) {
		if (typeof msg === "string") {
			if (!msg) {
				msg = 'NA';
			}
			else {
				if (shouldTrim === true) {
					msg = Logger.trimMsg(msg);
				}
				msg = msg.replace('|', '\\|');
			}
		}
		return msg;
	},
	
	/**
	 * convert object to escaped json to be displayed in the parameters section of a log line
	 * @param {object|string} params Object or JSON string to be outputed to the log
	 * @param {array} filter array of keys to filter the displayed params, or {key:.., value:..} for adding info not present in Obj
	 * @returns escaped json string
	 */
	prepParams: function(params, filter) {
		let outputObj = {};
		
		if (!params) {
			return "NA";
		}

		if (arguments.length == 2 && (params instanceof Object) && (filter instanceof Array)) {
			filter.forEach(function(key) {
				if (typeof key === 'string' && params.hasOwnProperty(key)) {
					outputObj[key] = Logger.escapeMsg(params[key], true);
				}
				else if (typeof key === 'object' && key["key"] && key["value"]) {
					outputObj[key["key"]] = Logger.escapeMsg(key["value"], true);
				}
				else {
					throw new TypeError('invalid parameter');
				}
			});
			return JSON.stringify(outputObj);
		}
		else if (arguments.length === 1 && (params instanceof Object)) {
			let keys = Object.keys(params);
			keys.forEach(function(key) {
				outputObj[key] = Logger.escapeMsg(params[key], true);
			});
			return JSON.stringify(outputObj);
		}
		else if (arguments.length === 1 && (typeof params === 'string')) {
			return Logger.escapeMsg(params);
		}
		else {
			throw new TypeError('invalid parameter');
		}
	},

	/**
	 * print log line
	 * @param {number} logLevel log level from Logger.levels
	 * @param {string} logModule module from Logger.modules
	 * @param {string} sessionId session id for tracking continuous operation across async calls
	 * @param {string} message log message
	 * @param {object|string} params log message parameters
	 */
	getLog: function(logLevel, logModule, sessionId, message, params) {
		try {
			Utils.assert(Logger.levelToText[logLevel], 'invalid log level');
			Utils.assert(logModule, 'invalid module');
			Utils.assert(message, 'invalid message');
			Utils.assert(sessionId, 'invalid session id');

			let format = [["LEVEL",Logger.levelToText[logLevel]],["MODULE",logModule], ["SESSION", sessionId], ["MESSAGE", '"'+message+'"'], ["PARAMS", Logger.prepParams(params)]];
			let log = format.map(function(v) {
				return v.join(': ');
			}).join(' | ');
			return log;
		}
		catch(e) {
			return "exception in logger: " + e.toString();
		}
	},

	// will ouput the message with the predefined format
	error: function(module, sessionId, message, params) { 
		if (Logger.levels.ERROR <= Logger.maxLogLevel) 
			error(Logger.getLog(Logger.levels.ERROR, module, sessionId, message, params)); 
	},
	warn: function(module, sessionId, message, params) {
		if (Logger.levels.WARN <= Logger.maxLogLevel) 
			warn(Logger.getLog(Logger.levels.WARN, module, sessionId, message, params)); 
	},
	info: function(module, sessionId, message, params) { 
		if (Logger.levels.INFO <= Logger.maxLogLevel)
			info(Logger.getLog(Logger.levels.INFO, module, sessionId, message, params));
	},
	debug: function(module, sessionId, message, params) { 
		if (Logger.levels.DEBUG <= Logger.maxLogLevel)
			debug(Logger.getLog(Logger.levels.DEBUG, module, sessionId, message, params)); 
	},
	automation: function(module, sessionId, message, params) {
		if (Logger.levels.AUTOMATION <= Logger.maxLogLevel)
			debug(Logger.getLog(Logger.levels.AUTOMATION, module, sessionId, message, params)); 
	},
	trace: function(module, sessionId, message, params) {
		if (Logger.levels.TRACE <= Logger.maxLogLevel)
			debug(Logger.getLog(Logger.levels.TRACE, module, sessionId, message, params));
	},

	// will ouput the message as given
	errorRaw: function(message) {
		if (Logger.levels.ERROR <= Logger.maxLogLevel) 
			error(message); 
	},
	warnRaw: function(message) {
		if (Logger.levels.WARN <= Logger.maxLogLevel) 
			warn(message); 
	},
	infoRaw: function(message) {
		if (Logger.levels.INFO <= Logger.maxLogLevel)
			info(message); 
	},
	debugRaw: function(message) {
		if (Logger.levels.DEBUG <= Logger.maxLogLevel)
			debug(message); 
	},
	automationRaw: function(message) { 
		if (Logger.levels.AUTOMATION <= Logger.maxLogLevel)
			debug(message); 
	},
	traceRaw: function(message) { 
		if (Logger.levels.TRACE <= Logger.maxLogLevel)
			debug(message); 
	}
};

// logger helper for download protection module
function logPrepFileParams(fileInfo) {
	return {
		"download_id": fileInfo["download_id"],
		"file_name": fileInfo["orig_file_name"],
		"url": fileInfo["url"],
		"content_type": fileInfo["fileData"]["content_type"]
	};
}

// logger helper for zero-phishing module
function logPrepZfParamsFromEntry(entry) {
	return {
		"domain": entry["domain"],
		"url": entry["url"]
	};
}

function collect_log_func(log_func) {
	return function (msg) {
		try {
			if (log_func)
				log_func(msg);
			if (collect_logs_enabled) {
				if (typeof msg !== 'string')
					msg = JSON.stringify(msg, undefined, "\t");
				logs_collections.push({ msg: msg, time: new Date() });
				if (logs_collections.length > 5000) {
					logs_collections.shift();
					if(AutomationLogger["new_logs_index"] > 0){ // for selenium automation
						AutomationLogger["new_logs_index"]--;
					}
				}
			}
		} catch (e) {
			console.log(e); // do not change it to warn or similar, this will cause infinite loop
		}
	};
}

function clear_collected_log() {
	logs_collections = [];
}

function print_collection_log() {
	for (var i = 0; i < logs_collections.length; i++)
		info(logs_collections[i].time._SBA4B_format('yyyy-MM-dd h:mm:ss:S') + ": " + logs_collections[i].msg);
}

function get_collected_log() {
	var collected_logs = "";
	for (var i = 0; i < logs_collections.length; i++)
		collected_logs += logs_collections[i].time._SBA4B_format('yyyy-MM-dd h:mm:ss:S') + ": " + logs_collections[i].msg + "\n";
	return collected_logs;
}

function getReverseMap(map) {
	var reverseMap = {};
    for (var prop in map) 
        if(map.hasOwnProperty(prop)) 
            reverseMap[map[prop]] = prop;
	return reverseMap;
}

function te_states(state) {
	switch (state) {
		case "background_emulation_finished":	
			return funcs.get_locale_message("LOCALE_SCANNING_FILE_FOR_THREATS");
		case "emulating":
			return funcs.get_locale_message("LOCALE_SCANNING_FILE_FOR_THREATS");
		case "uploading":
			return funcs.get_locale_message("LOCALE_SCANNING_FILE_FOR_THREATS");
		case "downloading":
			return funcs.get_locale_message("LOCALE_DOWNLOADING_FILE");
		case "finished":
			return funcs.get_locale_message("LOCALE_FINISHED_SCANNING_FILE"); 
	}
	return "";
}

function tex_states(state) {
	switch (state) {
		case "extracting":	
			return funcs.get_locale_message("LOCALE_SCANNING_FILE_FOR_THREATS");
		case "uploading":
			return funcs.get_locale_message("LOCALE_SCANNING_FILE_FOR_THREATS");
		case "downloading":
			return funcs.get_locale_message("LOCALE_DOWNLOADING_FILE");
		case "extracted_pdf":
			return funcs.get_locale_message("LOCALE_FILE_WAS_CONVERTED_TO_PDF_TO_REMOVE_SECURITY_RISKS");
		case "extracted_cleaned":
			return funcs.get_locale_message("LOCALE_POTENTIALLY_MALICIOUS_ELEMENTS_WERE_REMOVED_FROM_THIS_FILE"); 
	}
	return "";
}

function shouldExtract(/* int */ flags_per_file, /* int */ flags_per_family) 
{
	if (flags_per_file && flags_per_file !== 0)
		return ((flags_per_file & file_protection_mode_flags['extract']) !== 0);
	else
		return ((flags_per_family & file_protection_mode_flags['extract']) !== 0);
}

function shouldEmulate(/* int */ flags_per_file, /* int */ flags_per_family)
{
	if (flags_per_file && flags_per_file !== 0)
		return ((flags_per_file & file_protection_mode_flags['emulate']) !== 0);
	else
		return ((flags_per_family & file_protection_mode_flags['emulate']) !== 0);
}

function shouldEmulateInTheBackground(fileInfo)
{
	return (fileInfo["shouldEmulate"] && (fileInfo["threatEmulationMode"] == threat_emulation_flags["background"]));
}

function shouldAllow(/* int */ flags_per_file, /* int */ flags_per_family)
{
	if (flags_per_file && flags_per_file !== 0)
		return ((flags_per_file & file_protection_mode_flags['allow']) !== 0);
	else
		return ((flags_per_family & file_protection_mode_flags['allow']) !== 0);
}

function shouldBlock(/* int */ flags_per_file, /* int */ flags_per_family)
{
	if (flags_per_file && flags_per_file !== 0)
		return ((flags_per_file & file_protection_mode_flags['block']) !== 0);
	else
		return ((flags_per_family & file_protection_mode_flags['block']) !== 0);
}

function getFileProtectionMode(/* int */ flags)
{
	return (flags & 0x00F);
}

function getThreatEmulationMode(/* int */ flags_per_file, /* int */ flags_per_family)
{
	if (flags_per_file && flags_per_file !== 0)
		return ((flags_per_file & 0x0F0) >> 4);
	else
		return ((flags_per_family & 0x0F0) >> 4);
}

function getThreatExtractionMode(/* int */ flags_per_file, /* int */ flags_per_family)
{
	if (flags_per_file && flags_per_file !== 0)
		return ((flags_per_file & 0xF00) >> 8);
	else
		return ((flags_per_family & 0xF00) >> 8);
}

function calculateFileProtectionFlags(/* int */ mode, /* int */ te, /* int */ tex)
{
	return (mode | (te << 4) | (tex << 8));
}

function getFileExtension(file_name)
{
	if (file_name === undefined)
		return "";
	var lastIndex = file_name.lastIndexOf(".");
	if (lastIndex === -1)
		return "";
	return file_name.substr(lastIndex+1).toLowerCase();
}

function load_hooks() {
	load_hooks_impl();
	l();
	funcs.set_log_level(LOG_LVL_ERROR);	// comment this line if JS DEBUG level logs are needed
}

function l() {
	funcs.set_log_level(LOG_LVL_DEBUG);
}

function escapeStringForRegex(str) {
     return str.replace(/([.?*+^$[\]\\(){}|-])/g, "\\$1");
}

function isIP(domain) {
	try {
		var blocks = domain ? domain.split(".") : [];
		if (blocks.length === 4) {
			for (var i = 0; i < 4; i++) 	 
				if (!((parseInt(blocks[i], 10) >= 0) && (parseInt(blocks[i], 10) <= 255)))
					return false;
			return true;
		}
	} catch (e) {
		// swallow exception 
	}
	return false;
}

function isPrivateIP(ip) {
	if (!ip || !isIP(ip))
		return false;
	var parts = ip.split(".");
	return parts[0] === "10" || 
		(parts[0] === "172" && (parseInt(parts[1], 10) >= 16 && parseInt(parts[1], 10) <= 31)) || 
		(parts[0] === "192" && parts[1] === "168");
}

function getDomain(_url) {
	if (!_url || (_url == ""))
		return "";
	try {
		_url = _url.toLowerCase();
		if (_url.indexOf("blob:http") == 0)
			_url = _url.replace("blob:http", "http"); 
		if (is_ie_extension() && (_url.indexOf("blob:") == 0)) // For IE blob urls
			return "blob";
		var u = new unified_URL(_url);
		return u["hostname"] ? u["hostname"] : "";
	} catch (e) {
		// swallow exception 
	}
	return "";
}

function get_stripped_url(url) {
	if (!url || (url == ""))
		return "";
	var stripped_url = url;
	try {
		var u = new unified_URL(url.toLowerCase());
		if (!u["origin"]) 
			u["origin"] = u["protocol"] + "//" + u["hostname"] + ((u["port"] && (u["port"] != 80) && (u["port"] != 443)) ? (':' + u["port"]) : '');
		stripped_url = u["origin"] + u["pathname"];
		if (stripped_url._SBA4B_endsWith("/"))
			stripped_url = stripped_url.slice(0,-1);
	} catch (e) {
		// swallow exception
	}
	return decodeURIComponent(stripped_url);
}

function get_words(str) {
	str = str ? str : "";
	str = str.replace(/([a-z])([A-Z])/g, '$1 $2').replace(/\b([A-Z]+)([A-Z])([a-z])/, '$1 $2$3'); // handle camel case
	str = str.toLowerCase().replace(/([a-z])([0-9])/g, '$1 $2').replace(/([0-9])([a-z])/g, '$1 $2'); // handle numbers
	return str.split(/[-\.\s\/:_;~|{}\[\]()"',\?<>=+*&^%$#!`\\©Ⓒ]+/).filter(function(el) {return el.length != 0;} ); // check if we need to add @
}

function count_real_words(text) {
	var words = get_words(text);
	var ret = 0;
	for (var i = 0; i < words.length; i++)
		if ((words[i].trim().length > 2) && isNaN(words[i].trim()))
			ret++;
	return ret;
}

function get_lines(str) {
	var tmp = (str && str != "") ? str.trim().split(/\r\n|\r|\n/) : [];
	var ret = [];
	for (var i = 0; i < tmp.length; i++)
		if ((tmp[i].trim().length > 2) && (get_words(tmp[i]).length > 0))
			ret.push(tmp[i].toLowerCase());
	return ret;
}

// environment identification functions
function is_chrome_extension() {
	let browserName = funcs.browser_name();
	return (browserName === browser_name["chrome"]);
}

function is_firefox_quantum_extension() {
	return (funcs.browser_name() === browser_name["firefox_quantum"]);
}

function is_firefox_legacy_extension() {
	return (funcs.browser_name() === browser_name["firefox_legacy"]);
}

function is_chrome_or_firefox_quantum_extension(){
	return (is_chrome_extension() || is_firefox_quantum_extension());
}

function is_node() {
	return (funcs.browser_name() === browser_name["node"]);
}

function is_ie_extension() {
	return (funcs.browser_name() === browser_name["ie"]);
}

function is_edge_extension() {
	return (funcs.browser_name() === browser_name["edge"]);
}

function getRandomValues(arr) {
	for (var i=0; i< arr.length;i++ )
		arr[i]=Math.floor(Math.random() * 256);
}

//
function IPnumber(IPaddress) {
	var ip = IPaddress.match(/^(\d+)\.(\d+)\.(\d+)\.(\d+)$/);
	return ip ? ((+ip[1]<<24) + (+ip[2]<<16) + (+ip[3]<<8) + (+ip[4])) : 0;
}

function IPmask(maskSize) {
    return -1<<(32-maskSize);
}

// example: cidr 192.168.100/24, means subnet mask 255.255.255.0 => network addr is 192.168.100.0
function isIPInCIDR(ip, cidr) {
	if (!ip || !cidr)
		return false;
	var netMask = cidr ? cidr.split("/") : [];
	if ((netMask.length == 2) && !isNaN(netMask[1]))
		return (IPnumber(ip) & IPmask(netMask[1])) == (IPnumber(netMask[0]) & IPmask(netMask[1]));
	return false;
}

// will return true if:
// excluded domain is a suffix of a domain
// excluded domain is a regex, and that regex matches the domain
// excluded domain is a CIDR, the domain is an IP, and the network addresses match
function is_domain_in_domains(domain, domains, include_no_suffix_hosts) {
	if (!domain || (domain == "")) 
		return false;
	domain = domain.toLowerCase();
	var ip = isIP(domain);
	for (var i = 0; domains && (i < domains.length); i++) {
		try {
			var excluded_domain =  domains[i].toLowerCase();
			if (ip && isIPInCIDR(domain, excluded_domain)) {
				Logger.info(Logger.modules.MAIN, mainLogSessionId, "is_domain_in_domains(): domain excluded by cidr match",  { "domain": domain, "excluded_domian": excluded_domain });
				return true;
			}
				
			if (domain._SBA4B_endsWith(excluded_domain)) {
				Logger.info(Logger.modules.MAIN, mainLogSessionId, "is_domain_in_domains(): domain excluded by suffix match", { "domain": domain, "excluded_domian": excluded_domain });
				return true;
			}

			if (new RegExp("^" + excluded_domain + "$").test(domain)) {
				Logger.info(Logger.modules.MAIN, mainLogSessionId, "is_domain_in_domains(): domain excluded by regex match", { "domain": domain, "excluded_domian": excluded_domain });
				return true;
			}
		} catch (e) {
			Logger.error(Logger.modules.MAIN, mainLogSessionId, "exception in is_domain_in_domains()", { "exception": e });
		}
	}
	return include_no_suffix_hosts && (domain.indexOf(".") < 0) && (domain != "localhost");
}

function unified_addEventListener(elem,event,callback,optional){
	unified_removeEventListener(elem, event, callback, optional);
	if ( (typeof window != "undefined")  && (typeof window.addEventListener === "undefined")){
		if (event == "keypress") 
			document.attachEvent("on"+event, callback);
		else 
			elem.attachEvent("on" + event, callback);
	}else 
		elem.addEventListener(event,callback,optional);
}

function unified_removeEventListener(elem,event,callback,optional){
	if ( (typeof window != "undefined")  && (typeof window.addEventListener === "undefined")){
		if (event == "keypress") 
			document.detachEvent("on"+event, callback);
		else 
			elem.detachEvent("on" + event, callback);
	}else 
		elem.removeEventListener(event,callback,optional);
}

function binaryStringToBase64(buffer) {
    var binary = "";
    for (var i = 0; i < buffer.length; i++) 
        binary += String.fromCharCode(buffer.charCodeAt(i) & 0xff); // throw away high-order byte (f7)
    return btoa(binary);
}

function oldArrayBufferToBase64(buffer) {
	var binary = [];
    var bytes = new Uint8Array(buffer);
    for (var i = 0; i < bytes.byteLength; i++) 
        binary.push(String.fromCharCode(bytes[i]));
    return btoa(binary.join(""));
}

function arrayBufferToBase64(buffer, cb) {
	try {
		if (typeof FileReader == "undefined")
			return cb(oldArrayBufferToBase64(buffer));
		var blob = new Blob([buffer]);
		var fileReader =  new FileReader();
		fileReader.onload = function (evt) {
			try {
				cb(evt.target.result.split(",")[1]);
			} catch (e) {
				Logger.error(Logger.modules.MAIN, mainLogSessionId, "exception in array_buffer_to_base_64()", { "error": e });
				cb(oldArrayBufferToBase64(buffer));
			}
		};
		fileReader.onerror = function(e) {
			Logger.error(Logger.modules.MAIN, mainLogSessionId, "exception in array_buffer_to_base_64()", { "error": e });
			cb(oldArrayBufferToBase64(buffer));
		};
		fileReader.readAsDataURL(blob);
	} catch (e) {
		Logger.error(Logger.modules.MAIN, mainLogSessionId, "exception in array_buffer_to_base_64()", { "error": e });
		cb(oldArrayBufferToBase64(buffer));
	}
}

function dataURItoArray(dataURI) {
	// convert base64/URLEncoded data component to raw binary data held in a string
    var byteString;
	var dataURIArray = dataURI.split(',');
    if (dataURIArray[0].indexOf('base64') >= 0)
        byteString = atob(dataURIArray[1]);
    else
        byteString = unescape(dataURIArray[1]);

    // write the bytes of the string to a typed array
    var ia = new Uint8Array(byteString.length);
    for (var i = 0; i < byteString.length; i++) 
        ia[i] = byteString.charCodeAt(i);
	
	return ia;
}

function dataURItoBlob(dataURI) {
    var ia = dataURItoArray(dataURI);
	// separate out the mime component
    var mimeString = dataURI.split(',')[0].split(':')[1].split(';')[0];
    return new Blob([ia], {'type':mimeString});
}

/**
 * Cross browser implementation to save data as file
 * 
 * @param {String} fileName - Name of file to save
 * @param {String} data - Data to put in the file
 */
function saveAsFile(fileName, data) {
	var blob = new Blob([data], { 'type': 'application/octet-stream' });
	if (is_ie_extension()) {
		window.navigator.msSaveOrOpenBlob(blob, fileName);
	} else {
		var pom = document.createElement('a');		
		pom.href = unifiedCreateObjectURL(blob);
		pom.setAttribute('download', fileName);
		document.body.appendChild(pom);
		pom.click();	
	}
}

function get_locale_messageN(/*Array*/ locale_codes)
{
	var retVal = "";
	var arrayLength = locale_codes.length;
	for (var i = 0; i < arrayLength; i++) {
		retVal += funcs.get_locale_message(locale_codes[i]);
	}
	return retVal;
}

function localizeHTML() {
	$(".locale_message").each(function(i) {
		$(this).html(funcs.get_locale_message($(this).attr("data-locale-message-constnat")));
	});
}

/**
 * get the mime type from the value of a content type header
 * @param {string} contentType value of http content-type header, example: "text/html; charset=utf8"
 * @returns {string} mime type extracted from the content-type header, example "text/html"
 */
function getMimeFromContentType(contentType) {
	return contentType.split(";")[0].replace(/\s+/g,"").toLowerCase();
}

function is_assoative_array(obj){
	return obj && (typeof obj  === "object");
}

function IsDataUrl(url) {
	return (url.indexOf("data:") == 0);
}

function IsBlobUrl(url) {
	return (url.indexOf("blob:") == 0);
}

function b64_encode(text){
	return encodeURIComponent(btoa(unescape(encodeURIComponent(text))));
}

function equal_arrays(a, b){
	if (!a || !b || (a.length != b.length))
        return false;
	for (var i = 0; i < a.length; i++) 
		if (a[i] != b[i]) 
            return false;   
    return true;
}

/*
	Return True if the extension was deployed from a K12 distribution store id
*/
function is_K12_distribution() {
	if (is_chrome_extension()) {
		let distributionId = funcs.get_distribution_id();
		if (distributionId && K12_distributionIds[distributionId]) {
			Logger.info(Logger.modules.MAIN, mainLogSessionId, "is_K12_distribution(): this is a K12 distribution, detected by distribution id", { "destribution_id": distributionId });
			return true;
		}
		
		let manifest = funcs.get_manifest();
		if (manifest["name"] === "Check Point SandBlast Agent for Education") {
			Logger.info(Logger.modules.MAIN, mainLogSessionId, "is_K12_distribution(): this is a K12 distribution, detected by manifest");
			return true;
		}
	}
	return false;
}